/***************************************************************************
                    UNISIM TMS320C3X unit tests generator
                             -------------------
    begin                : Fri July 1st 2009
    copyright            : (C) 2009 Commissariat a l'Energie Atomique (CEA)
    authors              : Gilles Mouchard, Daniel Gracia Pérez
    email                : gilles.mouchard@cea.fr, daniel.gracia-perez@cea.fr
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <vector>
#include <queue>
#include <string>
#include <string.h>
#include <fstream>
#include <sstream>
#include <getopt.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <inttypes.h>

static const unsigned int ST_C = 0;
static const unsigned int ST_V = 1;
static const unsigned int ST_Z = 2;
static const unsigned int ST_N = 3;
static const unsigned int ST_UF = 4;
static const unsigned int ST_LV = 5;
static const unsigned int ST_LUF = 6;
static const unsigned int ST_OVM = 7;
static const unsigned int ST_RM = 8;
static const unsigned int ST_CF = 10;
static const unsigned int ST_CE = 11;
static const unsigned int ST_CC = 12;
static const unsigned int ST_GIE = 13;

static const unsigned int M_ST_C = 1 << ST_C;
static const unsigned int M_ST_V = 1 << ST_V;
static const unsigned int M_ST_Z = 1 << ST_Z;
static const unsigned int M_ST_N = 1 << ST_N;
static const unsigned int M_ST_UF = 1 << ST_UF;
static const unsigned int M_ST_LV = 1 << ST_LV;
static const unsigned int M_ST_LUF = 1 << ST_LUF;
static const unsigned int M_ST_OVM = 1 << ST_OVM;
static const unsigned int M_ST_RM = 1 << ST_RM;
static const unsigned int M_ST_CF = 1 << ST_CF;
static const unsigned int M_ST_CE = 1 << ST_CE;
static const unsigned int M_ST_CC = 1 << ST_CC;
static const unsigned int M_ST_GIE = 1 << ST_GIE;

using namespace std;

typedef enum
{
	NONE         = 0,
	INT_REG      = 1,
	FLOAT_REG    = 2,
	AUX_REG      = 3,
	ST           = 4,
	INT_BUFFER   = 5,
	FLOAT_BUFFER = 6
} operand_storage_type_t;

typedef enum
{
	SOURCE       = 1,
	DEST         = 2,
	SOURCE_DEST  = 3,
	REFERENCE    = 4,
	TEMPORARY    = 8,
	SUBSTITUTION = 16,
	CLOBBER      = 32
} operand_type_t;

static const unsigned int NUM_REFERENCES = 30;

struct keyword
{
	const char *name;
	operand_type_t type;
	operand_storage_type_t storage_type;
} keywords[] = {
	{ "src_int_reg",       SOURCE,       INT_REG },
	{ "src_float_reg",     SOURCE,       FLOAT_REG },
	{ "src_aux_reg",       SOURCE,       AUX_REG },
	{ "st_in",             SOURCE,       ST },
	{ "dst_int_reg",       DEST,         INT_REG },
	{ "dst_float_reg",     DEST,         FLOAT_REG },
	{ "dst_aux_reg",       DEST,         AUX_REG },
	{ "st_out",            DEST,         ST },
	{ "src_dst_int_reg",   SOURCE_DEST,  INT_REG },
	{ "src_dst_float_reg", SOURCE_DEST,  FLOAT_REG },
	{ "src_dst_aux_reg",   SOURCE_DEST,  AUX_REG },
	{ "tmp_int_reg",       TEMPORARY,    INT_REG },
	{ "tmp_float_reg",     TEMPORARY,    FLOAT_REG },
	{ "tmp_aux_reg",       TEMPORARY,    AUX_REG },
	{ "src_int_buf",       SOURCE,       INT_BUFFER },
	{ "dst_int_buf",       DEST,         INT_BUFFER },
	{ "src_float_buf",     SOURCE,       FLOAT_BUFFER },
	{ "dst_float_buf",     DEST,         FLOAT_BUFFER },
	{ "subst",             SUBSTITUTION, NONE },
	{ "clobber",           CLOBBER, NONE },
	{ "29",                REFERENCE,    NONE },
	{ "28",                REFERENCE,    NONE },
	{ "27",                REFERENCE,    NONE },
	{ "26",                REFERENCE,    NONE },
	{ "25",                REFERENCE,    NONE },
	{ "24",                REFERENCE,    NONE },
	{ "23",                REFERENCE,    NONE },
	{ "22",                REFERENCE,    NONE },
	{ "21",                REFERENCE,    NONE },
	{ "20",                REFERENCE,    NONE },
	{ "19",                REFERENCE,    NONE },
	{ "18",                REFERENCE,    NONE },
	{ "17",                REFERENCE,    NONE },
	{ "16",                REFERENCE,    NONE },
	{ "15",                REFERENCE,    NONE },
	{ "14",                REFERENCE,    NONE },
	{ "13",                REFERENCE,    NONE },
	{ "12",                REFERENCE,    NONE },
	{ "11",                REFERENCE,    NONE },
	{ "10",                REFERENCE,    NONE },
	{ "9",                 REFERENCE,    NONE },
	{ "8",                 REFERENCE,    NONE },
	{ "7",                 REFERENCE,    NONE },
	{ "6",                 REFERENCE,    NONE },
	{ "5",                 REFERENCE,    NONE },
	{ "4",                 REFERENCE,    NONE },
	{ "3",                 REFERENCE,    NONE },
	{ "2",                 REFERENCE,    NONE },
	{ "1",                 REFERENCE,    NONE },
	{ "0",                 REFERENCE,    NONE }
};

static const unsigned int REG_R0 = 0x00;
static const unsigned int REG_R1 = 0x01;
static const unsigned int REG_R2 = 0x02;
static const unsigned int REG_R3 = 0x03;
static const unsigned int REG_R4 = 0x04;
static const unsigned int REG_R5 = 0x05;
static const unsigned int REG_R6 = 0x06;
static const unsigned int REG_R7 = 0x07;
static const unsigned int REG_AR0 = 0x08;
static const unsigned int REG_AR1 = 0x09;
static const unsigned int REG_AR2 = 0x0a;
static const unsigned int REG_AR3 = 0x0b;
static const unsigned int REG_AR4 = 0x0c;
static const unsigned int REG_AR5 = 0x0d;
static const unsigned int REG_AR6 = 0x0e;
static const unsigned int REG_AR7 = 0x0f;
static const unsigned int REG_DP = 0x10;
static const unsigned int REG_IR0 = 0x11;
static const unsigned int REG_IR1 = 0x12;
static const unsigned int REG_BK = 0x13;
static const unsigned int REG_SP = 0x14;
static const unsigned int REG_ST = 0x15;
static const unsigned int REG_IE = 0x16;
static const unsigned int REG_IF = 0x17;
static const unsigned int REG_IOF = 0x18;
static const unsigned int REG_RS = 0x19;
static const unsigned int REG_RE = 0x1a;
static const unsigned int REG_RC = 0x1b;

static const char *reg_name[] = {
	"r0",  "r1",  "r2",  "r3",  "r4",  "r5",  "r6",  "r7",
	"ar0", "ar1", "ar2", "ar3", "ar4", "ar5", "ar6", "ar7",
	"dp", "ir0", "ir1",  "bk",  "sp",  "st",  "ie",  "if",
	"iof",  "rs",  "re",  "rc"
};

struct operand_t
{
	operand_type_t type;
	operand_storage_type_t storage_type;
	unsigned int pos_in_pattern;
	unsigned int length_in_pattern;
	unsigned int reg_num;
	unsigned int input_index;
	unsigned int output_index;
	unsigned int ref;
	unsigned int buffer_size;
	unsigned int subst_index;
	string subst_string;
	string clobbered_register_name;

	const char *get_substitution() const;
	const char *get_keyword() const;
	operand_type_t get_type() const;
	operand_storage_type_t get_storage_type() const;
	unsigned int get_ref() const;
	bool is_input() const;
	bool is_output() const;
};

class register_allocator_t
{
public:
	register_allocator_t();
	bool allocate(unsigned int& reg_num);
	void free(unsigned int reg_num);
	void reserve(unsigned int reg_num);
	bool is_register_allocated(unsigned int reg_num) const;
	bool is_register_reserved(unsigned int reg_num) const;
private:
	queue<unsigned int> free_list;
	unsigned int alloc_map;
	unsigned int res_map;
};

class unit_test_t
{
public:
	unit_test_t(const char *pattern);
	bool sanity_check() const;
	bool generate(ostream& os, const vector<string>& substitutions, bool pattern_under_test_only = false);
	const vector<operand_t>& get_operands() const;
	unsigned int get_input_dim() const;
	unsigned int get_output_dim() const;
	bool generate_random_input(ostream& os, unsigned int count);
private:
	bool parse_error;
	string pattern;
	vector<operand_t> operands;
	unsigned int input_dim;
	unsigned int output_dim;
	unsigned int num_substs;

	bool parse_buffer_subscript(const char *s, unsigned int& buffer_size, unsigned int& length) const;
	bool parse_clobber(const char *s, string& clobbered_register_name, unsigned int& length) const;
	bool parse_operand(const char *s, operand_type_t& type, operand_storage_type_t& storage_type, unsigned int& ref_index, unsigned int& length) const;
	const operand_t *find_operand(unsigned int pos_in_pattern) const;
	bool is_register_clobbered(unsigned int reg_num) const;
};

class testbench_t
{
public:
	static const unsigned int input_io_buffer_size = 4096;
	static const unsigned int output_io_buffer_size = 4096;
	testbench_t(unit_test_t *unit_test, const char *output_subdirname, const char *input_filename);
	bool generate(ostream& os);
private:
	unit_test_t *unit_test;
	string output_subdirname;
	string input_filename;
};

void latexify(ostream& os, const string& s, int line_width, int alignment_colno = 0)
{
	string end_of_line("\n");
	string alignment_token(";");
	int pos = 0;
	int lineno = 0;
	int colno = 1;
	bool line_break = false;
	bool new_line = false;
	bool alignment = false;
	bool not_blank = false;

	static struct
	{
		const char *searchFor;
		const char *replaceBy;
	} conversion_table[] = {
		"#", "\\#",
		" ", "~",
		"_", "\\_",
		"\t", "~~",
		"{", "$\\{$",
		"}", "$\\}$",
		"&", "\\&",
		"--", "{-}{-}",
		"<<", "<}\\texttt{<",
		"<--", "$\\leftarrow$",
		"<=", "$\\le$",
		">=", "$\\ge$",
		"<", "$<$",
		">", "$>$",
		"%", "\\%",
		";", ""
	};

	os << "\\texttt{";

	while(pos < s.length())
	{
		int i;
		int found = 0;

		if(new_line)
		{
			os << "}\\newline" << endl << "\\texttt{";
			new_line = false;
			not_blank = false;
		}

		if(line_break)
		{
			colno = 1;
			os << "}\\newline" << endl << "\\texttt{$\\hookrightarrow$~~~~";
		}

		if(alignment)
		{
			if(not_blank)
			{
				while(colno < alignment_colno)
				{
					colno++;
					os << "~";
				}
			}
			os << "\\it ";
		}
		
		line_break = colno > line_width && (s[pos] == ' ' || s[pos] == '\t' || s[pos] == '_' || s[pos] == '.');
		alignment = s.compare(pos, alignment_token.length(), alignment_token) == 0;
		not_blank = not_blank || (s[pos] != ';' && s[pos] != '\t' && s[pos] != ' ');

		for(i = 0; i < sizeof(conversion_table)/sizeof(conversion_table[0]); i++)
		{
			if(s.compare(pos, strlen(conversion_table[i].searchFor), conversion_table[i].searchFor) == 0)
			{
				os << conversion_table[i].replaceBy;
				pos += strlen(conversion_table[i].searchFor);
				colno += strlen(conversion_table[i].searchFor);
				found = 1;
				break;
			}
		}
		
		if(!found)
		{
			if(s.compare(pos, end_of_line.length(), end_of_line) == 0)
			{
				colno = 1;
				new_line = true;
				pos += end_of_line.length();
			}
			else
			{
				os << s[pos];
				pos++;
				colno++;
			}
		}
	}
	os << "}";
}

int count_lines_in_file(const char *filename)
{
	int n = 0;

	ifstream file(filename, ifstream::in);

	if(file.fail())
	{
		cout << "Can't open \"" << filename << "\"" << endl;
		return -1;
	}

	string s;
	while(1)
	{
		getline(file, s);
		if(file.fail()) break;
		n++;
	}

	return n;
}

class kiss_random_t
{
public:
	static const long Max = +2147483647L;
	static const long Min = -Max - 1L;
	kiss_random_t(int32_t seed_x, int32_t seed_y, int32_t seed_z, int32_t seed_w);
	int32_t generate();
	int32_t generate(uint32_t radius);
private:
	int32_t x;
	int32_t y;
	int32_t z;
	int32_t w;
	int32_t carry;
};

class kiss_probability_t
{
public:
	kiss_probability_t(int32_t seed_x, int32_t seed_y, int32_t seed_z, int32_t seed_w);
	void set_weight(unsigned int n, unsigned int weight);
	unsigned int generate();
private:
	kiss_random_t kiss_random;
	vector<unsigned int> distribution;
};

class kiss_random_ext_float_t
{
public:
	kiss_random_ext_float_t(int32_t seed_x, int32_t seed_y, int32_t seed_z, int32_t seed_w);
	void generate(uint32_t& exp, uint32_t &mant);
private:
	typedef enum
	{
		CAT_MINUS_INF    ,  // -inf
		CAT_SMALLEST_NEG ,  // smallest negative number
		CAT_ZERO         ,  // zero
		CAT_REAL_ZERO    ,  // real zero
		CAT_SMALLEST_POS ,  // smallest positive number
		CAT_INT			 ,  // near to int
		CAT_PLUS_INF     ,  // +inf
		CAT_SMALL        ,  // small
		CAT_LARGE        ,  // large
		CAT_NEAR_MANT    ,  // mantissa near previously generated, fully random exponent
		CAT_NEAR_EXP     , // exponent near previously generated, fully random mantissa
		CAT_NEAR         , // float near previously generated
		CAT_RANDOM         // fully random
	} Category;
	kiss_random_t kiss_random;
	kiss_probability_t proba;
	int32_t prev_mant;
	int32_t prev_exp;
};

class kiss_random_float_t
{
public:
	kiss_random_float_t(int32_t seed_x, int32_t seed_y, int32_t seed_z, int32_t seed_w);
	void generate(uint32_t &v);
private:
	typedef enum
	{
		CAT_MINUS_INF    ,  // -inf
		CAT_SMALLEST_NEG ,  // smallest negative number
		CAT_ZERO         ,  // zero
		CAT_REAL_ZERO	 ,  // real to zero
		CAT_SMALLEST_POS ,  // smallest positive number
		CAT_INT			 ,  // near to int
		CAT_PLUS_INF     ,  // +inf
		CAT_SMALL        ,  // small
		CAT_LARGE        ,  // large
		CAT_NEAR_MANT    ,  // mantissa near previously generated, fully random exponent
		CAT_NEAR_EXP     , // exponent near previously generated, fully random mantissa
		CAT_NEAR         , // float near previously generated
		CAT_RANDOM         // fully random
	} Category;
	kiss_random_t kiss_random;
	kiss_probability_t proba;
	int32_t prev_mant;
	int32_t prev_exp;
};

kiss_random_t::kiss_random_t(int32_t seed_x, int32_t seed_y, int32_t seed_z, int32_t seed_w)
	: x(seed_x)
	, y(seed_y)
	, z(seed_z)
	, w(seed_w)
	, carry(0)
{
}

int32_t kiss_random_t::generate()
{
	int32_t b;
	int32_t k;
	int32_t m;

	x = (69069 * x) + 1;
	y = y ^ (y << 13);
	y = y ^ (y >> 17);
	y = y ^ (y << 5);
	k = (z >> 2) + (w >> 3) + (carry >> 2);
	m = w + w + z + carry;
	z = w;
	w = m;
	carry = k >> 30;
	b = x + y + w;

	return b;
}

int32_t kiss_random_t::generate(uint32_t radius)
{
	// generate a random value between Min and Max
	int64_t r = generate();

	// compute current radius
	uint64_t cur_radius = (uint64_t) Max + 1;

	// scale
	r = r * radius / cur_radius;

	return (int32_t) r;
}

kiss_probability_t::kiss_probability_t(int32_t seed_x, int32_t seed_y, int32_t seed_z, int32_t seed_w)
	: kiss_random(seed_x, seed_y, seed_z, seed_w)
{
}

void kiss_probability_t::set_weight(unsigned int n, unsigned int weight)
{
	unsigned int i;
	for(i = 0; i < weight; i++)
	{
		distribution.push_back(n);
	}
}

unsigned int kiss_probability_t::generate()
{
	return distribution[((int64_t) kiss_random.generate() - (int64_t) kiss_random_t::Min) / (((int64_t) kiss_random_t::Max - (int64_t) kiss_random_t::Min + 1) / distribution.size())];
}

kiss_random_ext_float_t::kiss_random_ext_float_t(int32_t seed_x, int32_t seed_y, int32_t seed_z, int32_t seed_w)
	: kiss_random(seed_x, seed_y, seed_z, seed_w)
	, proba(seed_x, seed_y, seed_z, seed_w)
{
	proba.set_weight(CAT_MINUS_INF, 1);
	proba.set_weight(CAT_SMALLEST_NEG, 1);
	proba.set_weight(CAT_ZERO, 1);
	proba.set_weight(CAT_REAL_ZERO, 1);
	proba.set_weight(CAT_SMALLEST_POS, 1);
	proba.set_weight(CAT_PLUS_INF, 1);
	proba.set_weight(CAT_INT, 2);
	proba.set_weight(CAT_SMALL, 2);
	proba.set_weight(CAT_LARGE, 2);
	proba.set_weight(CAT_NEAR_MANT, 5);
	proba.set_weight(CAT_NEAR_EXP, 5);
	proba.set_weight(CAT_NEAR, 5);
	proba.set_weight(CAT_RANDOM, 10);
}

void kiss_random_ext_float_t::generate(uint32_t& exp, uint32_t& mant)
{
	int64_t tmp_exp = 0;
	int64_t tmp_mant = 0;

	switch(proba.generate())
	{
		case CAT_MINUS_INF:
			tmp_exp = 127;
			tmp_mant = -2147483647L - 1L;
			break;
		case CAT_SMALLEST_NEG:
			tmp_exp = -127;
			tmp_mant = -2147483647L - 1L;
			break;
		case CAT_ZERO:
			tmp_exp = -128;
			tmp_mant = kiss_random.generate();
			break;
		case CAT_REAL_ZERO:
			tmp_exp = -128;
			tmp_mant = 0;
			break;
		case CAT_SMALLEST_POS:
			tmp_exp = -127;
			tmp_mant = 0;
			break;
		case CAT_PLUS_INF:
			tmp_exp = +127;
			tmp_mant = +2147483647;
			break;
		case CAT_INT:
			tmp_exp = 0 + abs(kiss_random.generate(31));
			tmp_mant = kiss_random.generate();
			break;
		case CAT_SMALL:
			tmp_exp = -119 + kiss_random.generate(8);
			tmp_mant = kiss_random.generate();
			break;
		case CAT_LARGE:
			tmp_exp = 119 + kiss_random.generate(8);
			tmp_mant = kiss_random.generate();
			break;
		case CAT_NEAR_MANT:
			tmp_exp = kiss_random.generate(128);
			tmp_mant = prev_mant + kiss_random.generate(8);
			break;
		case CAT_NEAR_EXP:
			tmp_exp = prev_exp + kiss_random.generate(8);
			tmp_mant = kiss_random.generate();
			break;
		case CAT_NEAR:
			tmp_exp = prev_exp + kiss_random.generate(8);
			tmp_mant = prev_mant + kiss_random.generate(8);
			break;
		case CAT_RANDOM:
			tmp_exp = kiss_random.generate(128);
			tmp_mant = kiss_random.generate();
			break;
		default:
			assert(false);
			break;
	}

	if(tmp_exp < -128) tmp_exp = -128;
	if(tmp_exp > 127) tmp_exp = 127;
	if(tmp_mant > 2147483647) tmp_mant = +2147483647;
	if(tmp_mant < ( -2147483647L - 1L)) tmp_mant = -2147483647L - 1L;
	prev_exp = tmp_exp;
	prev_mant = tmp_mant;
	exp = (uint32_t) tmp_exp & 0xff;
	mant = (uint32_t) tmp_mant;
}

kiss_random_float_t::kiss_random_float_t(int32_t seed_x, int32_t seed_y, int32_t seed_z, int32_t seed_w)
	: kiss_random(seed_x, seed_y, seed_z, seed_w)
	, proba(seed_x, seed_y, seed_z, seed_w)
{
	proba.set_weight(CAT_MINUS_INF, 1);
	proba.set_weight(CAT_SMALLEST_NEG, 1);
	proba.set_weight(CAT_ZERO, 1);
	proba.set_weight(CAT_REAL_ZERO, 1);
	proba.set_weight(CAT_SMALLEST_POS, 1);
	proba.set_weight(CAT_PLUS_INF, 1);
	proba.set_weight(CAT_INT, 2);
	proba.set_weight(CAT_SMALL, 2);
	proba.set_weight(CAT_LARGE, 2);
	proba.set_weight(CAT_NEAR_MANT, 5);
	proba.set_weight(CAT_NEAR_EXP, 5);
	proba.set_weight(CAT_NEAR, 5);
	proba.set_weight(CAT_RANDOM, 10);
}

void kiss_random_float_t::generate(uint32_t& v)
{
	int64_t tmp_exp = 0;
	int64_t tmp_mant = 0;

	switch(proba.generate())
	{
		case CAT_MINUS_INF:
			tmp_exp = 127;
			tmp_mant = -8388608;
			break;
		case CAT_SMALLEST_NEG:
			tmp_exp = -127;
			tmp_mant = -8388608;
			break;
		case CAT_ZERO:
			tmp_exp = -128;
			tmp_mant = kiss_random.generate(8388608);
			break;
		case CAT_REAL_ZERO:
			tmp_exp = -128;
			tmp_mant = 0;
			break;
		case CAT_SMALLEST_POS:
			tmp_exp = -127;
			tmp_mant = 0;
			break;
		case CAT_PLUS_INF:
			tmp_exp = +127;
			tmp_mant = +8388607;
			break;
		case CAT_INT:
			tmp_exp = 0 + abs(kiss_random.generate(31));
			tmp_mant = kiss_random.generate(8388608);
			break;
		case CAT_SMALL:
			tmp_exp = -119 + kiss_random.generate(8);
			tmp_mant = kiss_random.generate(8388608);
			break;
		case CAT_LARGE:
			tmp_exp = 119 + kiss_random.generate(8);
			tmp_mant = kiss_random.generate(8388608);
			break;
		case CAT_NEAR_MANT:
			tmp_exp = kiss_random.generate(128);
			tmp_mant = prev_mant + kiss_random.generate(8);
			break;
		case CAT_NEAR_EXP:
			tmp_exp = prev_exp + kiss_random.generate(8);
			tmp_mant = kiss_random.generate(8388608);
			break;
		case CAT_NEAR:
			tmp_exp = prev_exp + kiss_random.generate(8);
			tmp_mant = prev_mant + kiss_random.generate(8);
			break;
		case CAT_RANDOM:
			tmp_exp = kiss_random.generate(128);
			tmp_mant = kiss_random.generate(8388608);
			break;
		default:
			assert(false);
			break;
	}

	if(tmp_exp < -128) tmp_exp = -128;
	if(tmp_exp > 127) tmp_exp = 127;
	if(tmp_mant < -8388608) tmp_mant = -8388608;
	if(tmp_mant > 8388607) tmp_mant = 8388607;
	prev_exp = tmp_exp;
	prev_mant = tmp_mant;
	v = ((uint32_t) (tmp_exp & 0xff) << 24) | ((uint32_t) tmp_mant & 0xffffff);
}

// Initialize random number generators
kiss_random_t kiss_random(-65475407, 654303921, -8453246, 745025667);
kiss_random_ext_float_t kiss_random_ext_float(-65475407, 654303921, -8453246, 745025667);
kiss_random_float_t kiss_random_float(-65475407, 654303921, -8453246, 745025667);

operand_type_t operand_t::get_type() const
{
	return type;
}

operand_storage_type_t operand_t::get_storage_type() const
{
	return storage_type;
}

unsigned int operand_t::get_ref() const
{
	return ref;
}

const char *operand_t::get_keyword() const
{
	unsigned int i;
	for(i = 0; i < (sizeof(keywords) / sizeof(keywords[0])); i++)
	{
		if(type == keywords[i].type && storage_type == keywords[i].storage_type) return keywords[i].name;
	}
	return 0;
}

const char *operand_t::get_substitution() const
{
	if(type & CLOBBER) return clobbered_register_name.c_str();
	if(type & SUBSTITUTION) return subst_string.c_str();
	return (reg_num != (unsigned int) - 1) ? reg_name[reg_num] : "???";
}

register_allocator_t::register_allocator_t()
	: free_list()
	, alloc_map(0)
	, res_map(0)
{
}

bool register_allocator_t::allocate(unsigned int& reg_num)
{
	if(free_list.empty()) return false;

	reg_num = free_list.front();
	if(res_map & (1 << reg_num)) return false;
	free_list.pop();
	alloc_map |= (1 << reg_num);
	return true;
}

void register_allocator_t::free(unsigned int reg_num)
{
	free_list.push(reg_num);
	alloc_map &= ~(1 << reg_num);
	res_map &= ~(1 << reg_num);
}

void register_allocator_t::reserve(unsigned int reg_num)
{
	res_map |= (1 << reg_num);
}

bool register_allocator_t::is_register_allocated(unsigned int reg_num) const
{
	return alloc_map & (1 << reg_num);
}

bool register_allocator_t::is_register_reserved(unsigned int reg_num) const
{
	return res_map & (1 << reg_num);
}

bool is_alpha_numeric(char c)
{
	return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9') || (c == '_');
}

bool is_numeric(char c)
{
	return c >= '0' && c <= '9';
}

bool is_separator(char c)
{
	return c == ' ' || c == '\t' || c == '\n';
}

bool is_register_name(const char *name)
{
	unsigned int reg_num;
	for(reg_num = 0; reg_num < (sizeof(reg_name) / sizeof(reg_name[0])); reg_num++)
	{
		if(strcmp(name, reg_name[reg_num]) == 0) return true;
	}
	return false;
}

bool unit_test_t::parse_buffer_subscript(const char *s, unsigned int& buffer_size, unsigned int& length) const
{
	const char *begin = s;
	buffer_size = 0;

	// skip spaces, tabs and ends of line
	while(is_separator(*s)) s++;

	if(*s++ != '[') return false;
	
	// skip spaces, tabs and ends of line
	while(is_separator(*s)) s++;

	if(!is_numeric(*s)) return false;

	// go to the end of buffer subscript
	while(is_numeric(*s))
	{
		buffer_size = (10 * buffer_size) + (*s - '0');
		s++;
	}

	// skip spaces, tabs and ends of line
	while(is_separator(*s)) s++;

	if(*s++ != ']') return false;

	length = s - begin;

	return true;
}

bool unit_test_t::parse_clobber(const char *s, string& clobbered_register_name, unsigned int& length) const
{
	const char *begin = s;
	length = 0;
	clobbered_register_name.clear();

	// skip spaces, tabs and ends of line
	while(is_separator(*s)) s++;

	// go to the end of buffer subscript
	while(is_alpha_numeric(*s))
	{
		clobbered_register_name.push_back(*s);
		s++;
	}

	if(!is_register_name(clobbered_register_name.c_str())) return false;

	length = s - begin;

	return true;
}


bool unit_test_t::parse_operand(const char *s, operand_type_t& type, operand_storage_type_t& storage_type, unsigned int& ref, unsigned int& length) const
{
	unsigned int i;
	for(i = 0; i < (sizeof(keywords) / sizeof(keywords[0])); i++)
	{
		length = strlen(keywords[i].name);

		if(strncmp(s, keywords[i].name, length) == 0 && !is_alpha_numeric(s[length])) 
		{
			type = keywords[i].type;
			storage_type = keywords[i].storage_type;
			// WARNING! last entries are for references
			ref = ((sizeof(keywords) / sizeof(keywords[0]) - 1 - i) < NUM_REFERENCES) ? (sizeof(keywords) / sizeof(keywords[0]) - 1 - i) : (unsigned int) -1;
			return true;
		}
	}
	return false;
}

unit_test_t::unit_test_t(const char *_pattern)
	: parse_error(false)
	, pattern(_pattern)
	, input_dim(0)
	, output_dim(0)
	, num_substs(0)
{
	unsigned int pos = 0;
	unsigned int lineno = 1;
	unsigned int colno = 1;
	unsigned int length;
	unsigned int operand_ref;
	operand_type_t operand_type;
	operand_storage_type_t operand_storage_type;
	unsigned int buffer_size;
	string clobbered_register_name;

	while(_pattern[pos])
	{
		if(_pattern[pos] == '%')
		{
			pos++;
			colno++;

			if(parse_operand(_pattern + pos, operand_type, operand_storage_type, operand_ref, length))
			{
				operand_t operand;

				operand.reg_num = (unsigned int) -1;
				operand.type = operand_type;
				operand.storage_type = operand_storage_type;
				operand.pos_in_pattern = pos;
				operand.length_in_pattern = length;
				operand.input_index = (unsigned int) -1;
				operand.output_index = (unsigned int) -1;
				operand.ref = (unsigned int) -1;
				operand.buffer_size = 0;
				operand.subst_index = (unsigned int) -1;
				operand.clobbered_register_name = "";

				if(operand_type & REFERENCE)
				{
					if(operand_ref >= operands.size())
					{
						cout << "Error! Undefined operand %" << operand_ref << " at line #" << lineno << " and column #" << colno << endl;
						parse_error = true;
						return;
					}
					operand.ref = operand_ref;
				}

				pos += length;
				colno += length;

				switch(operand.storage_type)
				{
					case INT_BUFFER:
					case FLOAT_BUFFER:
						if(!parse_buffer_subscript(_pattern + pos, buffer_size, length))
						{
							cout << "Parse Error! Expected buffer script at line #" << lineno << " and column #" << colno << endl;
							parse_error = true;
							return;
						}

						operand.buffer_size = buffer_size;
						operand.length_in_pattern += length;

						pos += length;
						colno += length;
						break;
					default:
						break;
				}

				if(operand.type & SOURCE)
				{
					operand.input_index = input_dim;

					switch(operand.storage_type)
					{
						case INT_REG:
						case ST:
						case AUX_REG:
							input_dim++;
							break;
						case FLOAT_REG:
							input_dim += 2;
							break;
						case INT_BUFFER:
						case FLOAT_BUFFER:
							input_dim += operand.buffer_size;
							break;
						default:
							assert(false);
							break;
					}
				}

				if(operand.type & DEST)
				{
					operand.output_index = output_dim;

					switch(operand.storage_type)
					{
						case INT_REG:
						case ST:
						case AUX_REG:
							output_dim++;
							break;
						case FLOAT_REG:
							output_dim += 2;
							break;
						case INT_BUFFER:
						case FLOAT_BUFFER:
							output_dim += operand.buffer_size;
							break;
						default:
							assert(false);
							break;
					}
				}

				if(operand.type & SUBSTITUTION)
				{
					operand.subst_index = num_substs++;
					operand.subst_string = string("");
				}

				if(operand_type & CLOBBER)
				{
					if(!is_separator(_pattern[pos]) || !parse_clobber(_pattern + pos, clobbered_register_name, length))
					{
						cout << "Parse Error! Expected clobbered register name after %clobber at line #" << lineno << " and column #" << colno << endl;
						parse_error = true;
						return;
					}

					operand.clobbered_register_name = clobbered_register_name;
					operand.length_in_pattern += length;

					pos += length;
					colno += length;
				}

				operands.push_back(operand);
			}
			else
			{
				if(_pattern[pos] == '%')
				{
					pos++;
					colno++;
				}
				else
				{
					cout << "Parse Error! Expected operand after % at line #" << lineno << " and column #" << colno << endl;
					parse_error = true;
					return;
				}
			}
		}
		else
		{
			if(_pattern[pos] == '\n')
			{
				lineno++;
				colno = 1;
			}
			pos++;
			colno++;
		}
	}
}

bool unit_test_t::sanity_check() const
{
	return !parse_error;
}

const vector<operand_t>& unit_test_t::get_operands() const
{
	return operands;
}

unsigned int unit_test_t::get_input_dim() const
{
	return input_dim;
}

unsigned int unit_test_t::get_output_dim() const
{
	return output_dim;
}

const operand_t *unit_test_t::find_operand(unsigned int pos_in_pattern) const
{
	vector<operand_t>::const_iterator operand_iter;

	for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
	{
		if(operand_iter->pos_in_pattern == pos_in_pattern) return &(*operand_iter);
	}

	return 0;
}

bool unit_test_t::is_register_clobbered(unsigned int reg_num) const
{
	vector<operand_t>::const_iterator operand_iter;

	for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
	{
		if(operand_iter->get_type() == CLOBBER && operand_iter->clobbered_register_name == reg_name[reg_num])
		{
			return true;
		}
	}

	return false;
}

bool unit_test_t::generate(ostream& os, const vector<string>& substitutions, bool pattern_under_test_only)
{
	vector<operand_t>::iterator operand_iter;
	vector<operand_t>::reverse_iterator operand_rev_iter;
	unsigned int tmp_reg_num;
	unsigned int input_addr_reg;
	unsigned int output_addr_reg;
	register_allocator_t reg_allocator;
	register_allocator_t aux_reg_allocator;

	int reg_num; // this is signed for a good reason

	// Fill-in register allocator
	for(reg_num = REG_R0; reg_num <= (int) REG_RC; reg_num++)
	{
		if(is_register_clobbered(reg_num))
		{
			reg_allocator.reserve(reg_num);
		}
	}

	for(reg_num = REG_R0; reg_num <= (int) REG_R7; reg_num++)
	{
		if(!reg_allocator.is_register_reserved(reg_num))
		{
			reg_allocator.free(reg_num);
		}
	}

	for(reg_num = REG_AR0; reg_num <= (int) REG_AR2; reg_num++)
	{
		if(!aux_reg_allocator.is_register_reserved(reg_num))
		{
			aux_reg_allocator.free(reg_num);
		}
	}

	// AR3 is reserved for frame pointer

	for(reg_num = REG_AR4; reg_num <= (int) REG_AR7; reg_num++)
	{
		if(!aux_reg_allocator.is_register_reserved(reg_num))
		{
			aux_reg_allocator.free(reg_num);
		}
	}

	// Allocate the registers
	if(!aux_reg_allocator.allocate(input_addr_reg) || !aux_reg_allocator.allocate(output_addr_reg))
	{
		cout << "Cannot allocate auxiliary register. Aborting!" << endl;
		return false;
	}

	for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
	{
		switch(operand_iter->get_storage_type())
		{
			case INT_REG:
			case ST:
			case FLOAT_REG:
				if(!reg_allocator.allocate(operand_iter->reg_num))
				{
					cout << "Cannot allocate register. Aborting!" << endl;
					return false;
				}
				break;
			case AUX_REG:
			case INT_BUFFER:
			case FLOAT_BUFFER:
				if(!aux_reg_allocator.allocate(operand_iter->reg_num))
				{
					cout << "Cannot allocate auxiliary register. Aborting!" << endl;
					return false;
				}
				break;
			case NONE:
				break;
		}
	}

	// Look at substitution strings
	for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
	{
		if(operand_iter->get_type() & SUBSTITUTION)
		{
			if(operand_iter->subst_index >= substitutions.size())
			{
				cerr << "ERROR! Not enough substitution strings" << endl;
				return false;
			}
			
			operand_iter->subst_string = substitutions[operand_iter->subst_index];
		}
	}

	if(!pattern_under_test_only)
	{
		// Start generating the assembly code
		os << "fp .set ar3" << endl;

		os << "\t.global _unit_test" << endl << endl;


		os << "_unit_test:" << endl;

		os << "\tpush fp" << endl;
		os << "\tldiu sp, fp" << endl;

		// Save the registers in forward order
		for(reg_num = 0; reg_num <= (int) REG_RC; reg_num++)
		{
			if(reg_allocator.is_register_allocated(reg_num) ||
			reg_allocator.is_register_reserved(reg_num) ||
			aux_reg_allocator.is_register_allocated(reg_num) ||
			aux_reg_allocator.is_register_reserved(reg_num))
			{
				os << "\tpush " << reg_name[reg_num] << endl;
			}
		}

		os << "\tldiu *-fp(2), " << reg_name[input_addr_reg] << " ; input" << endl;
		os << "\tldiu *-fp(3), " << reg_name[output_addr_reg] << " ; output" << endl;

		// Read input array
		for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
		{
			if(operand_iter->get_type() & SOURCE)
			{
				switch(operand_iter->get_storage_type())
				{
					case INT_REG:
					case AUX_REG:
					case ST:
						os << "\tldiu *+" << reg_name[input_addr_reg] << "(" << operand_iter->input_index << "), " << operand_iter->get_substitution() << endl;
						break;
					case INT_BUFFER:
					case FLOAT_BUFFER:
						os << "\tldiu " << reg_name[input_addr_reg] << ", " << operand_iter->get_substitution();
						if(operand_iter->input_index)
						{
							os << endl << "\taddi " << operand_iter->input_index << ", " << operand_iter->get_substitution();
						}
						os << "  ; address of an input buffer of size " << operand_iter->buffer_size << endl;
						break;
					case FLOAT_REG:
						if(!reg_allocator.allocate(tmp_reg_num))
						{
							cout << "Cannot allocate temporary register. Aborting!" << endl;
							return false;
						}

						os << "\tpush " << reg_name[tmp_reg_num] << endl;
						os << "\tldiu *+" << reg_name[input_addr_reg] << "(" << operand_iter->input_index << "), " << reg_name[tmp_reg_num] << endl;
						os << "\tlsh 24, " << reg_name[tmp_reg_num] << endl;
						os << "\tpush " << reg_name[tmp_reg_num] << endl;
						os << "\tpopf " << operand_iter->get_substitution() << endl;
						os << "\tldm *+" << reg_name[input_addr_reg] << "(" << (operand_iter->input_index + 1) << "), " << operand_iter->get_substitution() << "   ; " << operand_iter->get_substitution() << " <- (exp[7-0] || mant[31-0])" << endl;
						os << "\tpop " << reg_name[tmp_reg_num] << endl;

						reg_allocator.free(tmp_reg_num);
						break;
					default:
						assert(false);
						break;
				}
			}

			if(operand_iter->get_type() & DEST)
			{
				switch(operand_iter->get_storage_type())
				{
					case INT_REG:
					case AUX_REG:
					case ST:
						break;
					case INT_BUFFER:
					case FLOAT_BUFFER:
						os << "\tldiu " << reg_name[output_addr_reg] << ", " << operand_iter->get_substitution();
						if(operand_iter->output_index)
						{
							os << endl << "\taddi " << operand_iter->output_index << ", " << operand_iter->get_substitution();
						}
						os << "  ; address of an output buffer of size " << operand_iter->buffer_size << endl;
						break;
					case FLOAT_REG:
						break;
					default:
						assert(false);
						break;
				}
			}
		}

		// Generate instruction pattern by substituing all the operands
		os << "\t; --- instruction pattern under test ---" << endl;
	}

	os << ";---- INPUTS ----" << endl;

	for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
	{
		if(operand_iter->get_type() & SOURCE && operand_iter->reg_num != (unsigned int) -1)
		{
			os << "; " << reg_name[operand_iter->reg_num] << ": ";

			switch(operand_iter->get_storage_type())
			{
				case NONE: break;
				case INT_REG: os << "a 32-bit integer register"; break;
				case FLOAT_REG: os << "a 40-bit floating point register"; break;
				case AUX_REG: os << "an auxiliary register"; break;
				case ST: os << "a 32-bit integer register (value for st)"; break;
				case INT_BUFFER: os << "an auxiliary register pointing to an array of " << operand_iter->buffer_size << " 32-bit integer values"; break;
				case FLOAT_BUFFER: os << "an auxiliary register pointing to an array of " << operand_iter->buffer_size << " 32-bit floating point values"; break;
			}

			os << endl;
		}
	}

	os << ";---- OUTPUTS ----" << endl;

	for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
	{
		if(operand_iter->get_type() & DEST && operand_iter->reg_num != (unsigned int) -1)
		{
			os << "; " << reg_name[operand_iter->reg_num] << ": ";

			switch(operand_iter->get_storage_type())
			{
				case NONE: break;
				case INT_REG: os << "a 32-bit integer register"; break;
				case FLOAT_REG: os << "a 40-bit floating point register"; break;
				case AUX_REG: os << "an auxiliary register"; break;
				case ST: os << "a 32-bit integer register (value of st)"; break;
				case INT_BUFFER: os << "an auxiliary register pointing to an array of " << operand_iter->buffer_size << " 32-bit integer values"; break;
				case FLOAT_BUFFER: os << "an auxiliary register pointing to an array of " << operand_iter->buffer_size << " 32-bit floating point values"; break;
			}

			os << endl;
		}
	}

	unsigned int pos = 0;

	while(pattern[pos])
	{
		if(pattern[pos] == '%')
		{
			pos++;

			if(pattern[pos] != '%')
			{
				const operand_t *operand;

				operand = find_operand(pos);

				if(!operand)
				{
					cout << "Internal error" << endl;
					return false;
				}

				unsigned int ref = operand->get_ref();

				if(ref != (unsigned int) -1)
				{
					unsigned int cur = ref;
					unsigned int prev;

					do
					{
						prev = cur;
						cur = operands[cur].ref;
					} while(cur != (unsigned int) -1);

					os << operands[prev].get_substitution();
				}
				else
				{
					os << operand->get_substitution();
				}

				pos += operand->length_in_pattern;
			}
			else
			{
				os << pattern[pos];
				pos++;
			}
		}
		else
		{
			os << pattern[pos];
			pos++;
		}
	}

	if(pos > 0 && pattern[pos - 1] != '\n')
	{
		// insert a newline at the end of the pattern if it's missing
		os << endl;
	}

	if(!pattern_under_test_only)
	{
		os << "\t; --------------------------------------" << endl;

		// Fill-in output array
		for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
		{
			if(operand_iter->get_type() & DEST)
			{
				switch(operand_iter->get_storage_type())
				{
					case INT_REG:
					case AUX_REG:
					case ST:
						os << "\tsti " << operand_iter->get_substitution() << ", *+" << reg_name[output_addr_reg] << "(" << operand_iter->output_index << ")" << endl;
						break;
					case FLOAT_REG:
						if(!reg_allocator.allocate(tmp_reg_num))
						{
							cout << "Cannot allocate temporary register. Aborting!" << endl;
							return false;
						}
						
						os << "\tpush " << reg_name[tmp_reg_num] << endl;
						os << "\tpushf " << operand_iter->get_substitution() << endl;
						os << "\tpop " << reg_name[tmp_reg_num] << endl;
						os << "\tlsh -24, " << reg_name[tmp_reg_num] << endl;
						os << "\tsti " << reg_name[tmp_reg_num] << ", *+" << reg_name[output_addr_reg] << "(" << operand_iter->output_index << ")  ; exponent of result" << endl;
						os << "\tsti " << operand_iter->get_substitution() << ", *+" << reg_name[output_addr_reg] << "(" << (operand_iter->output_index + 1) << ")  ; mantissa of result" << endl;
						os << "\tpop " << reg_name[tmp_reg_num] << endl;

						reg_allocator.free(tmp_reg_num);
						break;
					case INT_BUFFER:
					case FLOAT_BUFFER:
						break;
					default:
						assert(false);
						break;
				}
			}
		}

		// Restore the registers in reverse order
		for(reg_num = REG_RC; reg_num >= 0; reg_num--)
		{
			if(reg_allocator.is_register_allocated(reg_num) ||
			reg_allocator.is_register_reserved(reg_num) ||
			aux_reg_allocator.is_register_allocated(reg_num) ||
			aux_reg_allocator.is_register_reserved(reg_num))
			{
				os << "\tpop " << reg_name[reg_num] << endl;
			}
		}

		os << "\tpop fp" << endl;
		os << "\tretsu" << endl;
	}

	return true;
}

bool unit_test_t::generate_random_input(ostream& os, unsigned int count)
{
	while(count > 0)
	{
		vector<operand_t>::const_iterator operand_iter;
		unsigned int i;
		uint32_t exp;
		uint32_t mant;
		uint32_t v;
		bool first;

		for(first = true, operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
		{
			if(operand_iter->get_type() & SOURCE)
			{
				if(first)
				{
					first = false;
				}
				else
				{
					os << " ";
				}
				switch(operand_iter->get_storage_type())
				{
					case INT_REG:
					case AUX_REG:
						v = kiss_random.generate();
						os << std::hex << v << std::dec;
						break;
					case ST:
						v = kiss_random.generate() & ~M_ST_RM; // ST[RM]=0, other bits are randomized;
						os << std::hex << v << std::dec;
						break;
					case INT_BUFFER:
						for(i = 0; i < operand_iter->buffer_size; i++)
						{
							if(i > 0) os << " ";
							v = kiss_random.generate();
							os << std::hex << v << std::dec;
						}
						break;
					case FLOAT_BUFFER:
						for(i = 0; i < operand_iter->buffer_size; i++)
						{
							if(i > 0) os << " ";
							kiss_random_float.generate(v);
							os << std::hex << v << std::dec;
						}
						break;
					case FLOAT_REG:
						kiss_random_ext_float.generate(exp, mant);
						os << std::hex << exp << " " << mant << std::dec;
						break;
					default:
						assert(false);
						break;
				}
			}
		}
		os << endl;

		count--;
	}

	return true;
}

testbench_t::testbench_t(unit_test_t *_unit_test, const char *_output_subdirname, const char *_input_filename)
	: unit_test(_unit_test)
	, output_subdirname(_output_subdirname)
	, input_filename(_input_filename)
{
}

bool testbench_t::generate(ostream& os)
{
	unsigned int i;
	vector<operand_t>::const_iterator operand_iter;
	const vector<operand_t>& operands = unit_test->get_operands();
	unsigned int input_dim = unit_test->get_input_dim();
	unsigned int output_dim = unit_test->get_output_dim();

	if(output_dim == 0)
	{
		cerr << "ERROR! Unit test has no outputs" << endl;
		return false;
	}

	os << "#include <stdio.h>" << endl;
	os << "#include <stdlib.h>" << endl;
	os << "extern enable_insn_cache();" << endl;
	os << "extern void unit_test(unsigned int *input, unsigned int *output);" << endl;
	os << "#define OUTPUT_FILENAME \"" << output_subdirname << "/output.txt\"" << endl << endl;
	os << "#define INPUT_IO_BUFFER_SIZE " << input_io_buffer_size << endl;
	os << "#define OUTPUT_IO_BUFFER_SIZE " << output_io_buffer_size << endl;
	os << "#define ST_C 0" << endl;
	os << "#define ST_V 1" << endl;
	os << "#define ST_Z 2" << endl;
	os << "#define ST_N 3" << endl;
	os << "#define ST_UF 4" << endl;
	os << "#define ST_LV 5" << endl;
	os << "#define ST_LUF 6" << endl;
	os << "#define ST_OVM 7" << endl;
	os << "#define ST_RM 8" << endl;
	os << "#define ST_CF 10" << endl;
	os << "#define ST_CE 11" << endl;
	os << "#define ST_CC 12" << endl;
	os << "#define ST_GIE 13" << endl << endl;

	os << "void main()" << endl;
	os << "{" << endl;
	os << "\tconst char *INPUT_FILENAMES[2] = { \"" << input_filename << "\", \"" << output_subdirname << "/random.txt\" };" << endl;
	os << "\tFILE *fin;" << endl;
	os << "\tFILE *fout;" << endl;

	if(input_dim > 0)
	{
		os << "\tunsigned int input[" << input_dim << "];" << endl;
	}
	else
	{
		os << "\tunsigned int *input = 0; /* no inputs */" << endl;
	}
	os << "\tunsigned int output[" << output_dim << "];" << endl;
	os << "\tunsigned int i, j, k = 0;" << endl;
	os << "\tenable_insn_cache();" << endl;

	os << "\tprintf(\"Opening file \\\"%s\\\"\\n\", OUTPUT_FILENAME);" << endl;

	os << "\tfout = fopen(OUTPUT_FILENAME, \"wt\");" << endl;

	os << "\tif(!fout)" << endl;
	os << "\t{" << endl;
	os << "\t\tfprintf(stderr, \"Can't open output file \\\"%s\\\"\\n\", OUTPUT_FILENAME);" << endl;
	os << "\t\texit(-1);" << endl;
	os << "\t}" << endl;

	os << "\tif(setvbuf(fout, 0, _IOFBF, OUTPUT_IO_BUFFER_SIZE) != 0)" << endl;
	os << "\t{" << endl;
	os << "\t\tfprintf(stderr, \"WARNING! Can't increase output I/O buffer size\");" << endl;
	os << "\t}" << endl;
	os << "\tfor(j = 0; j < sizeof(INPUT_FILENAMES) / sizeof(INPUT_FILENAMES[0]); j++)" << endl;
	os << "\t{" << endl;
	os << "\t\tprintf(\"Opening file \\\"%s\\\"\\n\", INPUT_FILENAMES[j]);" << endl;

	os << "\t\tfin = fopen(INPUT_FILENAMES[j], \"rt\");" << endl;

	os << "\t\tif(!fin)" << endl;
	os << "\t\t{" << endl;
	os << "\t\t\tfprintf(stderr, \"Can't open input file \\\"%s\\\"\\n\", INPUT_FILENAMES[j]);" << endl;
	os << "\t\t\texit(-1);" << endl;
	os << "\t\t}" << endl;

	os << "\t\tif(setvbuf(fin, 0, _IOFBF, INPUT_IO_BUFFER_SIZE) != 0)" << endl;
	os << "\t\t{" << endl;
	os << "\t\t\tfprintf(stderr, \"WARNING! Can't increase output I/O buffer size\");" << endl;
	os << "\t\t}" << endl;
	os << "\t\tprintf(\"Starting tests\\n\");" << endl;
	os << "\t\tfprintf(fout, \"== %s ==\\n\", INPUT_FILENAMES[j]);" << endl;

	if(input_dim > 0)
	{
		os << "\t\ti = 1;" << endl;
		os << "\t\twhile(fscanf(fin, \"";

		for(i = 0; i < input_dim; i++)
		{
			os << "%x" << ((i == (input_dim - 1)) ? "\\n" : " ");
		}

		os << "\", ";

		for(i = 0; i < input_dim; i++)
		{
			os << "&input[" << i << "]" << ((i == (input_dim - 1)) ? ")" : ", ");
		}

		os << " == " << input_dim << ")" << endl;
		os << "\t\t{" << endl;
		os << "\t\t\tfprintf(fout, \"#%u\\n\", i);" << endl;
		os << "\t\t\tif(i >= k) { printf(\"#%u\\n\", i); k += 100; }" << endl;

		os << "\t\t\t/* print inputs */" << endl;
		os << "\t\t\tfprintf(fout, \"";

		for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
		{
			if(operand_iter->get_type() & SOURCE)
			{
				os << "\\t";
				switch(operand_iter->get_storage_type())
				{
					case INT_REG:
						os << "src_int_reg=%08x";
						break;
					case AUX_REG:
						os << "src_aux_reg=%08x";
						break;
					case FLOAT_REG:
						os << "src_float_reg=%02x %08x";
						break;
					case ST:
						os << "st_in=%08x (C=%u V=%u Z=%u N=%u UF=%u LV=%u LUF=%u OVM=%u RM=%u CF=%u CE=%u CC=%u GIE=%u)";
						break;
					case INT_BUFFER:
					case FLOAT_BUFFER:
						os << "src_buf=";
						for(i = 0; i < operand_iter->buffer_size; i++)
						{
							os << "%08x";
							if(i != (operand_iter->buffer_size - 1)) os << " ";
						}
						break;
					default:
						assert(false);
						break;
				}
				
				os << "\\n";
			}
		}

		os << "\"";

		for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
		{
			if(operand_iter->get_type() & SOURCE)
			{
				os << ", ";
				switch(operand_iter->get_storage_type())
				{
					case INT_REG:
						os << "input[" << operand_iter->input_index << "]";
						break;
					case AUX_REG:
						os << "input[" << operand_iter->input_index << "]";
						break;
					case FLOAT_REG:
						os << "input[" << operand_iter->input_index << "], input[" << (operand_iter->input_index + 1) << "]";
						break;
					case ST:
						os << "input[" << operand_iter->input_index << "], ";
						os << "(input[" << operand_iter->input_index  << "] >> ST_C) & 1, (input[" << operand_iter->input_index  << "] >> ST_V) & 1, (input[" << operand_iter->input_index  << "] >> ST_Z) & 1, (input[" << operand_iter->input_index  << "] >> ST_N) & 1, (input[" << operand_iter->input_index  << "] >> ST_UF) & 1, (input[" << operand_iter->input_index  << "] >> ST_LV) & 1, (input[" << operand_iter->input_index  << "] >> ST_LUF) & 1, (input[" << operand_iter->input_index  << "] >> ST_OVM) & 1, (input[" << operand_iter->input_index  << "] >> ST_RM) & 1, (input[" << operand_iter->input_index  << "] >> ST_CF) & 1, (input[" << operand_iter->input_index  << "] >> ST_CE) & 1, (input[" << operand_iter->input_index  << "] >> ST_CC) & 1, (input[" << operand_iter->input_index  << "] >> ST_GIE) & 1";
						break;
					case INT_BUFFER:
					case FLOAT_BUFFER:
						for(i = 0; i < operand_iter->buffer_size; i++)
						{
							os << "input[" << (operand_iter->input_index + i) << "]";
							if(i != (operand_iter->buffer_size - 1)) os << ", ";
						}
						break;
					default:
						assert(false);
						break;
				}
			}
		}

		os << ");" << endl;
	}

	os << "\t\t\t/* clearing output */" << endl;
	os << "\t\t\tmemset(output, 0, sizeof(output));" << endl;
	os << "\t\t\t/* call function containing instruction pattern under test */" << endl;
	os << "\t\t\t" << "unit_test(input, output);" << endl;

	os << "\t\t\t/* print outputs */" << endl;
	os << "\t\t\tfprintf(fout, \"";

	for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
	{
		if(operand_iter->get_type() & DEST)
		{
			os << "\\t";
			switch(operand_iter->get_storage_type())
			{
				case INT_REG:
					os << "dst_int_reg=%08x";
					break;
				case AUX_REG:
					os << "dst_aux_reg=%08x";
					break;
				case FLOAT_REG:
					os << "dst_float_reg=%02x %08x";
					break;
				case ST:
					os << "st_out=%08x (C=%u V=%u Z=%u N=%u UF=%u LV=%u LUF=%u OVM=%u RM=%u CF=%u CE=%u CC=%u GIE=%u)";
					break;
				case INT_BUFFER:
				case FLOAT_BUFFER:
					os << "dst_buf=";
					for(i = 0; i < operand_iter->buffer_size; i++)
					{
						os << "%08x";
						if(i != (operand_iter->buffer_size - 1)) os << " ";
					}
					break;
				default:
					assert(false);
					break;
			}
			
			os << "\\n";
		}
	}

	os << "\"";

	for(operand_iter = operands.begin(); operand_iter != operands.end(); operand_iter++)
	{
		if(operand_iter->get_type() & DEST)
		{
			os << ", ";
			switch(operand_iter->get_storage_type())
			{
				case INT_REG:
					os << "output[" << operand_iter->output_index << "]";
					break;
				case AUX_REG:
					os << "output[" << operand_iter->output_index << "]";
					break;
				case FLOAT_REG:
					os << "output[" << operand_iter->output_index << "], output[" << (operand_iter->output_index + 1) << "]";
					break;
				case ST:
					os << "output[" << operand_iter->output_index << "], ";
					os << "(output[" << operand_iter->output_index  << "] >> ST_C) & 1, (output[" << operand_iter->output_index  << "] >> ST_V) & 1, (output[" << operand_iter->output_index  << "] >> ST_Z) & 1, (output[" << operand_iter->output_index  << "] >> ST_N) & 1, (output[" << operand_iter->output_index  << "] >> ST_UF) & 1, (output[" << operand_iter->output_index  << "] >> ST_LV) & 1, (output[" << operand_iter->output_index  << "] >> ST_LUF) & 1, (output[" << operand_iter->output_index  << "] >> ST_OVM) & 1, (output[" << operand_iter->output_index  << "] >> ST_RM) & 1, (output[" << operand_iter->output_index  << "] >> ST_CF) & 1, (output[" << operand_iter->output_index  << "] >> ST_CE) & 1, (output[" << operand_iter->output_index  << "] >> ST_CC) & 1, (output[" << operand_iter->output_index  << "] >> ST_GIE) & 1";
					break;
				case INT_BUFFER:
				case FLOAT_BUFFER:
					for(i = 0; i < operand_iter->buffer_size; i++)
					{
						os << "output[" << (operand_iter->output_index + i) << "]";
						if(i != (operand_iter->buffer_size - 1)) os << ", ";
					}
					break;
				default:
					assert(false);
					break;
			}
		}
	}

	os << ");" << endl;

	if(input_dim > 0)
	{
		os << "\t\t\ti++;" << endl;
		os << "\t\t}" << endl;
	}
	os << "\t\tfclose(fin);" << endl;
	os << "\t}" << endl;
	os << "\tfclose(fout);" << endl;

	os << "\tprintf(\"Bye!\\n\");" << endl;
	os << "\texit(0);" << endl;
	os << "}" << endl;

	return true;
}

string string_to_identifier(const char *s)
{
	string ident;

	if(is_numeric(*s)) ident.push_back('_');

	while(*s)
	{
		if(is_alpha_numeric(*s))
			ident.push_back(*s);
		else
			ident.push_back('_');

		s++;
	}

	return ident;
}

void mkdir_p(const char *dirname)
{
	
}

void help(char *prog_name)
{
	cerr << "Description:" << endl;
	cerr <<  "   This is a unit tests generator for the UNISIM TMS320C3X simulator." << endl << endl;
	cerr << "Usage: " << prog_name << " <options> [substitutions...]" << endl << endl;
	cerr << "    'substitutions' are the strings replacing each %subst in order of appearance in the pattern" << endl; 
	cerr << "    At least one of the following options must be used: --random, --unittest, --testbench or --doc" << endl;
	cerr << "    The following options are mandatory: --pattern, --output" << endl;
	cerr << "    If --testbench is used then --input is mandatory" << endl;
	cerr << "    If --doc or --random is used then --random-count is mandatory" << endl;
	cerr << "    If --doc is used then --input is mandatory" << endl;
	cerr << "Options:" << endl;
	cerr << "  --help" << endl;
	cerr << "  -h" << endl;
	cerr << "                                    print this help" << endl << endl;
	cerr << "  --pattern <pattern filename>" << endl;
	cerr << "  -p <pattern filename>" << endl;
	cerr << "                                    set the pattern filename to use" << endl << endl;
	cerr << "  --input <input filename>" << endl;
	cerr << "  -i <input filename>" << endl;
	cerr << "                                    set the input data set file to use" << endl << endl;
	cerr << "  --output <sub-directory name>" << endl;
	cerr << "  -o <sub-directory name>" << endl;
	cerr << "                                    set the sub-directory name" << endl << endl;
	cerr << "  --random" << endl;
	cerr << "  -r" << endl;
	cerr << "                                    generate random unit test inputs" << endl << endl;
	cerr << "  --unittest" << endl;
	cerr << "  -u" << endl;
	cerr << "                                    generate a unit test in test.asm" << endl << endl;
	cerr << "  --testbench" << endl;
	cerr << "  -t" << endl;
	cerr << "                                    generate a testbench in main.c" << endl << endl;
	cerr << "  --doc" << endl;
	cerr << "  -d" << endl;
	cerr << "                                    generate unit test documentation (Latex file)" << endl << endl;
	cerr << "  --random-count <count>" << endl;
	cerr << "  -c <count>" << endl;
	cerr << "                                    set the number of random unit test inputs" << endl << endl;
	cerr << "  --force" << endl;
	cerr << "  -f" << endl;
	cerr << "                                    force overwriting existing files" << endl << endl;
}

void Copyright()
{
	cerr << "UNISIM TMS320C3X unit tests generator" << endl;
	cerr << "Copyright (c) 2009, Commissariat a l'Energie Atomique (CEA)" << endl;
	cerr << "Authors: Gilles Mouchard <gilles.mouchard@cea.fr>" << endl;
	cerr << "         Daniel Gracia Pérez <daniel.gracia-perez@cea.fr>" << endl;
	cerr << "License: GNU General Public License version 2, see file COPYING for details" << endl << endl;
}

int main(int argc, char *argv[])
{
	int i;
	int c;
	const char *pattern_filename = 0;
	const char *input_filename = 0;
	const char *output_subdirname = 0;
	bool has_random = false;
	bool has_random_count = false;
	unsigned int random_count = 0;
	bool has_testbench = false;
	bool has_unittest = false;
	bool has_doc = false;
	bool force_overwriting = false;

	static struct option long_options[] = 
	{
		{"help", no_argument, 0, 'h'},
		{"pattern", required_argument, 0, 'p'},
		{"input", required_argument, 0, 'i'},
		{"outdir", required_argument, 0, 'o'},
		{"random", no_argument, 0, 'r'},
		{"testbench", no_argument, 0, 't'},
		{"unittest", no_argument, 0, 'u'},
		{"doc", no_argument, 0, 'd'},
		{"random-count", required_argument, 0, 'c'},
		{"force", no_argument, 0, 'f'},
		{0, 0, 0, 0}
	};

	char line[1024];

	while((c = getopt_long(argc, argv, "hp:i:o:s:rtudfc:", long_options, 0)) != -1)
	{
		switch(c)
		{
			case 'h':
				help(argv[0]);
				return -1;
				break;
			case 'p':
				pattern_filename = optarg;
				break;
			case 'i':
				input_filename = optarg;
				break;
			case 'o':
				output_subdirname = optarg;
				break;
			case 'r':
				has_random = true;
				break;
			case 'c':
				random_count = atoi(optarg);
				has_random_count = true;
				break;
			case 't':
				has_testbench = true;
				break;
			case 'u':
				has_unittest = true;
				break;
			case 'd':
				has_doc = true;
				break;
			case 'f':
				force_overwriting = true;
				break;
		}
	}

	if(!pattern_filename ||
	   (has_testbench && !input_filename) ||
	   !output_subdirname ||
	   (!has_random && !has_testbench && !has_unittest && !has_doc) ||
	   (has_doc && (!input_filename || !has_random_count)) ||
	   (has_random && !has_random_count))
	{
		Copyright();
		help(argv[0]);
		return -1;
	}

	// get the substition strings
	vector<string> substitutions;
	for(i = optind; i < argc; i++)
		substitutions.push_back(argv[i]);

	cout << "Opening \"" << pattern_filename << "\"" << endl;
	ifstream pattern_file(pattern_filename, ifstream::in);

	if(pattern_file.fail())
	{
		cout << "Can't open \"" << pattern_filename << "\"" << endl;
		return -1;
	}

	string pattern;

	while(!pattern_file.getline(line, sizeof(line)).fail())
	{
		pattern += line;
		pattern += '\n';
	}

	unit_test_t unit_test(pattern.c_str());

	if(!unit_test.sanity_check())
	{
		cout << "Generation aborted" << endl;
		return -1;
	}

	if(has_unittest)
	{
		string unit_test_source_filename(string(output_subdirname) + string("/test.asm"));

		if(force_overwriting || access(unit_test_source_filename.c_str(), F_OK) != 0)
		{
			cout << "Creating \"" << unit_test_source_filename << "\"" << endl;
			ofstream unit_test_source_file(unit_test_source_filename.c_str(), ifstream::out);

			if(unit_test_source_file.fail())
			{
				cout << "Can't create \"" << unit_test_source_filename << "\"" << endl;
				return -1;
			}

			if(!unit_test.generate(unit_test_source_file, substitutions))
			{
				cout << "Generation aborted" << endl;
				unit_test_source_file.close();
				unlink(unit_test_source_filename.c_str());
				return -1;
			}
		}
	}

	if(has_random)
	{
		string random_input_filename(string(output_subdirname) + string("/random.txt"));

		if(force_overwriting || access(random_input_filename.c_str(), F_OK) != 0)
		{
			cout << "Creating \"" << random_input_filename << "\"" << endl;
			ofstream random_input_file(random_input_filename.c_str(), ifstream::out);

			if(random_input_file.fail())
			{
				cout << "Can't create \"" << random_input_filename << "\"" << endl;
				return -1;
			}

			if(!unit_test.generate_random_input(random_input_file, random_count))
			{
				cout << "Generation aborted" << endl;
				random_input_file.close();
				unlink(random_input_filename.c_str());
				return -1;
			}
		}
	}

	if(has_testbench)
	{
		string testbench_source_filename(string(output_subdirname) + string("/main.c"));

		if(force_overwriting || access(testbench_source_filename.c_str(), F_OK) != 0)
		{
			cout << "Creating \"" << testbench_source_filename << "\"" << endl;
			ofstream testbench_source_file(testbench_source_filename.c_str(), ifstream::out);

			if(testbench_source_file.fail())
			{
				cout << "Can't create \"" << testbench_source_filename << "\"" << endl;
				return -1;
			}

			testbench_t testbench(&unit_test, output_subdirname, input_filename);

			if(!testbench.generate(testbench_source_file))
			{
				cout << "Generation aborted" << endl;
				testbench_source_file.close();
				unlink(testbench_source_filename.c_str());
				return -1;
			}
		}
	}

	if(has_doc)
	{
		string doc_filename(string(output_subdirname) + string("/test.tex"));

		if(force_overwriting || access(doc_filename.c_str(), F_OK) != 0)
		{
			cout << "Creating \"" << doc_filename << "\"" << endl;
			ofstream doc_file(doc_filename.c_str(), ifstream::out);

			if(doc_file.fail())
			{
				cout << "Can't create \"" << doc_filename << "\"" << endl;
				return -1;
			}

			string random_input_filename(string(output_subdirname) + string("/random.txt"));
			stringstream sstr;

			if(!unit_test.generate(sstr, substitutions, true /* pattern under test only */))
			{
				cout << "Generation of documentation aborted" << endl;
				doc_file.close();
				unlink(doc_filename.c_str());
				return -1;
			}

			doc_file << "\\label{" << string_to_identifier(output_subdirname) << "}" << endl;
			doc_file << "\\begin{center}" << endl;
			doc_file << "\\begin{tabular}{|p{17cm}|}" << endl;
			doc_file << "\\hline" << endl;
			doc_file << "\\multicolumn{1}{|p{17cm}|}{\\textbf{Subdirectory}:~";
			latexify(doc_file, output_subdirname, 80);
			doc_file << "}\\\\" << endl;
			doc_file << "\\multicolumn{1}{|p{17cm}|}{\\textbf{Pattern}:~";
			latexify(doc_file, pattern_filename, 85);
			doc_file << "}\\\\" << endl;
			doc_file << "\\multicolumn{1}{|p{17cm}|}{\\textbf{Manually selected input}:~";
			latexify(doc_file, input_filename, 67);
			int number_of_inputs = count_lines_in_file(input_filename);
			if(number_of_inputs < 0)
			{
				cout << "Generation of documentation aborted" << endl;
				doc_file.close();
				unlink(doc_filename.c_str());
				return -1;
			}
			doc_file << "~~("<< number_of_inputs << " tests)}\\\\" << endl;
			doc_file << "\\multicolumn{1}{|p{17cm}|}{\\textbf{Random input}:~";
			latexify(doc_file, random_input_filename, 90);
			doc_file << "~~("<< random_count << " tests)}\\\\" << endl;
			doc_file << "\\multicolumn{1}{|p{17cm}|}{\\textbf{Assembly pattern under test}:\\newline";
			latexify(doc_file, sstr.str(), 130, 30);
			doc_file << "}\\\\" << endl;
			doc_file << "\\hline" << endl;
			doc_file << "\\end{tabular}" << endl;
			doc_file << "\\end{center}" << endl;
		}
	}
	return 0;
}
